// To be included in sched.c

#include <linux/bossa_policy_rts.h>

// unblocking

static inline int
rts_unblock(int caller, struct task_struct * p, int synchronous) {
  unsigned long flags;

  /* In Bossa, we only notify an unblock event if the process is not
   * running or in the ready queue In Linux we may have useless/multiple
   * calls to wake_up which may cause the Bossa run-time system to notify
   * an unblock event more than once (an unneeded/inexisting unblock event)
   * */

  p->state = TASK_RUNNING;
  if (synchronous) {
    int res = bossa_check_process_state(&(p->bossa),BLOCKED);
    bossa_notify_event(EVENT_UNBLOCK_NONPREEMPTIVE|caller,(&(p->bossa)),NULL);
    return res;
  }
  else {
    int res;
    spin_lock_irqsave(&bossa_scheduler_lock, flags);
#ifdef CONFIG_BOSSA_LIBRARY
    pre_check_unblock_preemptive(&p->bossa);
#endif
    res = root_scheduler.ops->unblock_preemptive_entry(EVENT_UNBLOCK_PREEMPTIVE|caller,(&(p->bossa)));
#ifdef CONFIG_BOSSA_LIBRARY
    post_check();
    root_s_state = compute_scheduler_state();
#endif
    spin_unlock_irqrestore(&bossa_scheduler_lock, flags);
    // the following should be in the policy...
    if (res && root_s_state == READY) {
      current->need_resched = 1;
    }
    return res;
  }
}


// scheduling functions

#define attr(x) list_entry(x, struct task_struct, bossa)

static inline void yield_system_immediate(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_yield_schedule(running);
#endif
  root_s_state =
    root_scheduler.ops->yield_system_immediate_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void yield_system_pause(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_yield_schedule(running);
#endif
  root_s_state =
    root_scheduler.ops->yield_system_pause_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void block(int caller) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_block(running);
#endif
  root_s_state = root_scheduler.ops->block_entry(caller,running);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void bossa_schedule(void) {
#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_bossa_schedule();
#endif
  root_s_state = root_scheduler.ops->bossa_schedule_entry();
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif
}

static inline void __schedule_tail(struct task_struct *prev);
asmlinkage inline void rts_schedule(int caller)
{
	struct schedule_data * sched_data;
	struct task_struct *prev, *next;
	int this_cpu;
	struct bossa_struct *local_old_running;

	if (!current->active_mm) BUG();
need_resched_back:
	prev = current;
	this_cpu = prev->processor;

	if (unlikely(in_interrupt())) {
		printk("Scheduling in interrupt\n");
		BUG();
	}

	release_kernel_lock(prev, this_cpu);

	/*
	 * 'sched_data' is protected by the fact that we can run
	 * only one process per CPU.
	 */
	sched_data = & aligned_data[this_cpu].schedule_data;
	  
	/* Bossa RTS begins here */
	spin_lock_irq(&bossa_scheduler_lock);
	local_old_running = running;

	if (running == rts_idle_process) {
	  bossa_schedule();
	}
	else {
	  switch (prev->state) {
	  case TASK_INTERRUPTIBLE:
	    if (!signal_pending(prev)) {
	      block(caller);
	      prev->policy &= ~SCHED_YIELD;
	      break;
	    } // fall through
	    prev->state = TASK_RUNNING;
	  case TASK_RUNNING:
	    if (prev->policy & SCHED_YIELD) {
	      prev->policy &= ~SCHED_YIELD;
	      yield_system_pause(caller); // root state can't be blocked
	      break;
	    }
	    else {
	      switch (root_s_state) {
	      case RUNNING:
		yield_system_immediate(caller);
		break;
	      case READY:
		bossa_schedule();
		break;
	      case BLOCKED:
	      default:
		running = rts_idle_process;
		break;
	      }
	      break;
	    }
	  case TASK_UNINTERRUPTIBLE:
	  case TASK_STOPPED:
	  case TASK_ZOMBIE:
	    block(caller);
	    break;
	  }
	}
	prev->need_resched = 0;
	spin_unlock_irq(&bossa_scheduler_lock);

	if (running == local_old_running) {
	  goto outbossa;
	}

#ifdef CONFIG_BOSSA_TRACE
	get_jinfo();
#endif

	// We are here only if there is a new process to run

	// Warning switch_to is not a proc.. prev and next are mandatory          
	prev = attr(local_old_running);
	next = attr(running);

	/*
	 * from this point on nothing can prevent us from
	 * switching to the next task, save this fact in
	 * sched_data.
	 */
	sched_data->curr = next;
	kstat.context_swtch++;
	/*
	 * there are 3 processes which are affected by a context switch:
	 *
	 * prev == .... ==> (last => next)
	 *
	 * It's the 'much more previous' 'prev' that is on next's stack,
	 * but prev is set to (the just run) 'last' process by switch_to().
	 * This might sound slightly confusing but makes tons of sense.
	 */

	prepare_to_switch();
	{
		struct mm_struct *mm = next->mm;
		struct mm_struct *oldmm = prev->active_mm;
		if (!mm) {
			if (next->active_mm) BUG();
			next->active_mm = oldmm;
			atomic_inc(&oldmm->mm_count);
			enter_lazy_tlb(oldmm, next, this_cpu);
		} else {
			if (next->active_mm != mm) BUG();
			switch_mm(oldmm, mm, next, this_cpu);
		}

		if (!prev->mm) {
			prev->active_mm = NULL;
			mmdrop(oldmm);
		}
	}

	/*
	 * This just switches the register state and the
	 * stack.
	 */

	switch_to(prev, next, prev);
	__schedule_tail(prev);

outbossa:
	reacquire_kernel_lock(current);
	// we check the bossa state, because that could have changed after
	// the new process is chosen, even though current is not yet
	// updated to the new process at that point
	// anyway, the state seems to be always READY when need_resched
	// is set, except in init/main.c, which is not relevant here
	if (root_s_state == READY /*current->need_resched*/) {
		goto need_resched_back;
	}
	return;
}

static void process_timeout(unsigned long __data);

signed long inline rts_schedule_timeout(int caller, signed long timeout)
{
	struct timer_list timer;
	unsigned long expire;

	switch (timeout)
	{
	case MAX_SCHEDULE_TIMEOUT:
		/*
		 * These two special cases are useful to be comfortable
		 * in the caller. Nothing more. We could take
		 * MAX_SCHEDULE_TIMEOUT from one of the negative value
		 * but I' d like to return a valid offset (>=0) to allow
		 * the caller to do everything it want with the retval.
		 */
		{
		rts_schedule(caller); // timeout too high, set no timer.
		goto out;
		}
	default:
		/*
		 * Another bit of PARANOID. Note that the retval will be
		 * 0 since no piece of kernel is supposed to do a check
		 * for a negative retval of schedule_timeout() (since it
		 * should never happens anyway). You just have the printk()
		 * that will tell you if something is gone wrong and where.
		 */
		if (timeout < 0)
		{
			printk(KERN_ERR "schedule_timeout: wrong timeout "
			       "value %lx from %p\n", timeout,
			       __builtin_return_address(0));
			current->state = TASK_RUNNING;
			goto out;
		}
	}

	expire = timeout + jiffies;

	init_timer(&timer);
	timer.expires = expire;
	timer.data = (unsigned long) current;
	timer.function = process_timeout;

	add_timer(&timer);
	 // change event name -> block_wait_queue _no_signal ?
        rts_schedule(caller);
	del_timer_sync(&timer);

	timeout = expire - jiffies;

 out:
	return timeout < 0 ? 0 : timeout;
}

signed long schedule_timeout(signed long timeout) {
  return rts_schedule_timeout(0x08888,timeout);
}

asmlinkage void schedule(void) {
  rts_schedule(0x07777);
}

asmlinkage void schedule_from_entry(void) {
  rts_schedule(0x09999);
}
